import csv
import os
import tempfile
from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference
from pyproteininference import in_silico_digest
from pyproteininference.parameters import ProteinInferenceParameter

TEST_DATABASE = resource_filename("pyproteininference", "../tests/data/test_database.fasta")
TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_exclusion.yaml")
OUTPUT_DIR = tempfile.gettempdir()
# OUTPUT_DIR = resource_filename('pyproteininference', '../tests/output/')
for sub_dir in ["leads", "all", "peptides", "psms", "psm_ids"]:
    if not os.path.exists(os.path.join(OUTPUT_DIR, sub_dir)):
        os.makedirs(os.path.join(OUTPUT_DIR, sub_dir))

LEAD_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/leads/test_exclusion_q_value_leads_ml_posterior_error_prob.csv",
)
ALL_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/all/test_exclusion_q_value_all_ml_posterior_error_prob.csv",
)
PEPTIDE_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/peptides/test_exclusion_q_value_leads_peptides_ml_posterior_error_prob.csv",
)
PSM_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/psms/test_exclusion_q_value_leads_psms_ml_posterior_error_prob.csv",
)
PSM_ID_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/psm_ids/test_exclusion_q_value_leads_psm_ids_ml_posterior_error_prob.csv",
)

IDENTIFIER_INDEX = 0
SCORE_INDEX = 1
Q_VALUE_INDEX = 2
GROUP_ID_INDEX = 5
PEPTIDES_INDEX = 6


class TestLoadExclusionWorkflow(TestCase):
    def test_workflow_exclusion(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        self.assertEqual(protein_inference_parameters.digest_type, "trypsin")
        self.assertEqual(protein_inference_parameters.export, "q_value")
        self.assertEqual(protein_inference_parameters.fdr, 0.01)
        self.assertEqual(protein_inference_parameters.missed_cleavages, 3)
        self.assertEqual(protein_inference_parameters.picker, True)
        self.assertEqual(protein_inference_parameters.restrict_pep, 0.9)
        self.assertEqual(protein_inference_parameters.restrict_peptide_length, 7)
        self.assertEqual(protein_inference_parameters.restrict_q, 0.9)
        self.assertEqual(protein_inference_parameters.protein_score, "multiplicative_log")
        self.assertEqual(protein_inference_parameters.psm_score, "posterior_error_prob")
        self.assertEqual(protein_inference_parameters.psm_score_type, "multiplicative")
        self.assertEqual(protein_inference_parameters.decoy_symbol, "##")
        self.assertEqual(protein_inference_parameters.isoform_symbol, "-")
        self.assertEqual(protein_inference_parameters.reviewed_identifier_symbol, "sp|")
        self.assertEqual(protein_inference_parameters.inference_type, "exclusion")
        self.assertEqual(protein_inference_parameters.tag, "test_exclusion")
        self.assertEqual(protein_inference_parameters.grouping_type, None)
        self.assertEqual(protein_inference_parameters.max_identifiers_peptide_centric, 5)
        self.assertEqual(protein_inference_parameters.lp_solver, None)

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=False,
        )
        pep_and_prot_data.read_psms()

        self.assertEqual(len(pep_and_prot_data.psms), 27)

        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        data.restrict_psm_data()

        self.assertEqual(len(data.main_data_restricted), 26)

        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        data.create_scoring_input()

        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        if protein_inference_parameters.inference_type == pyproteininference.inference.Inference.EXCLUSION:
            # This gets ran if we run exclusion...
            data.exclude_non_distinguishing_peptides()

        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        score = pyproteininference.scoring.Score(data=data)
        score.score_psms(score_method=protein_inference_parameters.protein_score)

        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        if protein_inference_parameters.picker:
            data.protein_picker()
        else:
            pass

        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        inference_type = protein_inference_parameters.inference_type

        if inference_type == pyproteininference.inference.Inference.PARSIMONY:
            group = pyproteininference.inference.Parsimony(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.INCLUSION:
            group = pyproteininference.inference.Inclusion(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.EXCLUSION:
            group = pyproteininference.inference.Exclusion(data=data, digest=digest)
            group.infer_proteins()

        # STEP 11: Run FDR and Q value Calculations
        # STEP 11: Run FDR and Q value Calculations
        # STEP 11: Run FDR and Q value Calculations
        data.calculate_q_values()

        # STEP 12: Export to CSV
        # STEP 12: Export to CSV
        # STEP 12: Export to CSV
        export_type = protein_inference_parameters.export
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=os.path.join(OUTPUT_DIR, "leads"), export_type=export_type)

        lead_output = []
        with open(LEAD_OUTPUT_FILE, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                lead_output.append(row)

        del lead_output[0]

        protein_groups = data.protein_group_objects

        for i in range(len(protein_groups)):
            lead_protein = protein_groups[i].proteins[0]
            self.assertEqual(lead_protein.identifier, lead_output[i][IDENTIFIER_INDEX])
            self.assertAlmostEqual(lead_protein.score, float(lead_output[i][SCORE_INDEX]))
            self.assertEqual(protein_groups[i].q_value, float(lead_output[i][Q_VALUE_INDEX]))
            self.assertEqual(protein_groups[i].number_id, int(lead_output[i][GROUP_ID_INDEX]))
            self.assertEqual(lead_protein.peptides, set(lead_output[i][PEPTIDES_INDEX:]))

        export.export_to_csv(directory=os.path.join(OUTPUT_DIR, "all"), export_type="q_value_all")

        all_output = []
        with open(ALL_OUTPUT_FILE, "r") as all_output_file:
            reader = csv.reader(all_output_file, delimiter=",")
            for row in reader:
                all_output.append(row)

        del all_output[0]

        all_output_new = []
        with open(export.filepath, "r") as all_output_file_new:
            reader = csv.reader(all_output_file_new, delimiter=",")
            for row in reader:
                all_output_new.append(row)

        del all_output_new[0]

        for i in range(len(all_output)):
            self.assertEqual(all_output_new[i][IDENTIFIER_INDEX], all_output[i][IDENTIFIER_INDEX])
            self.assertAlmostEqual(float(all_output_new[i][SCORE_INDEX]), float(all_output[i][SCORE_INDEX]))
            self.assertEqual(
                float(all_output_new[i][Q_VALUE_INDEX]),
                float(all_output[i][Q_VALUE_INDEX]),
            )
            self.assertEqual(
                int(all_output_new[i][GROUP_ID_INDEX]),
                int(all_output[i][GROUP_ID_INDEX]),
            )
            self.assertEqual(
                set(all_output_new[i][PEPTIDES_INDEX:]),
                set(all_output[i][PEPTIDES_INDEX:]),
            )

        export.export_to_csv(directory=os.path.join(OUTPUT_DIR, "peptides"), export_type="peptides")

        peptide_output = []
        with open(PEPTIDE_OUTPUT_FILE, "r") as peptide_output_file:
            reader = csv.reader(peptide_output_file, delimiter=",")
            for row in reader:
                peptide_output.append(row)

        del peptide_output[0]

        peptide_output_new = []
        with open(export.filepath, "r") as peptide_output_file_new:
            reader = csv.reader(peptide_output_file_new, delimiter=",")
            for row in reader:
                peptide_output_new.append(row)

        del peptide_output_new[0]

        for i in range(len(peptide_output)):
            self.assertEqual(
                peptide_output_new[i][IDENTIFIER_INDEX],
                peptide_output[i][IDENTIFIER_INDEX],
            )
            self.assertAlmostEqual(
                float(peptide_output_new[i][SCORE_INDEX]),
                float(peptide_output[i][SCORE_INDEX]),
            )
            self.assertEqual(
                float(peptide_output_new[i][Q_VALUE_INDEX]),
                float(peptide_output[i][Q_VALUE_INDEX]),
            )
            self.assertEqual(
                int(peptide_output_new[i][GROUP_ID_INDEX]),
                int(peptide_output[i][GROUP_ID_INDEX]),
            )
            self.assertEqual(
                set(peptide_output_new[i][PEPTIDES_INDEX:]),
                set(peptide_output[i][PEPTIDES_INDEX:]),
            )

        export.export_to_csv(directory=os.path.join(OUTPUT_DIR, "psms"), export_type="psms")

        psm_output = []
        with open(PSM_OUTPUT_FILE, "r") as psm_output_file:
            reader = csv.reader(psm_output_file, delimiter=",")
            for row in reader:
                psm_output.append(row)

        del psm_output[0]

        psm_output_new = []
        with open(export.filepath, "r") as psm_output_file_new:
            reader = csv.reader(psm_output_file_new, delimiter=",")
            for row in reader:
                psm_output_new.append(row)

        del psm_output_new[0]

        for i in range(len(psm_output)):
            self.assertEqual(psm_output_new[i][IDENTIFIER_INDEX], psm_output[i][IDENTIFIER_INDEX])
            self.assertAlmostEqual(float(psm_output_new[i][SCORE_INDEX]), float(psm_output[i][SCORE_INDEX]))
            self.assertEqual(
                float(psm_output_new[i][Q_VALUE_INDEX]),
                float(psm_output[i][Q_VALUE_INDEX]),
            )
            self.assertEqual(
                int(psm_output_new[i][GROUP_ID_INDEX]),
                int(psm_output[i][GROUP_ID_INDEX]),
            )
            self.assertEqual(
                set(psm_output_new[i][PEPTIDES_INDEX:]),
                set(psm_output[i][PEPTIDES_INDEX:]),
            )

        export.export_to_csv(directory=os.path.join(OUTPUT_DIR, "psm_ids"), export_type="psm_ids")

        psm_id_output = []
        with open(PSM_ID_OUTPUT_FILE, "r") as psm_id_output_file:
            reader = csv.reader(psm_id_output_file, delimiter=",")
            for row in reader:
                psm_id_output.append(row)

        del psm_id_output[0]

        psm_id_output_new = []
        with open(export.filepath, "r") as psm_id_output_file_new:
            reader = csv.reader(psm_id_output_file_new, delimiter=",")
            for row in reader:
                psm_id_output_new.append(row)

        del psm_id_output_new[0]

        for i in range(len(psm_id_output)):
            self.assertEqual(
                psm_id_output_new[i][IDENTIFIER_INDEX],
                psm_id_output[i][IDENTIFIER_INDEX],
            )
            self.assertAlmostEqual(
                float(psm_id_output_new[i][SCORE_INDEX]),
                float(psm_id_output[i][SCORE_INDEX]),
            )
            self.assertEqual(
                float(psm_id_output_new[i][Q_VALUE_INDEX]),
                float(psm_id_output[i][Q_VALUE_INDEX]),
            )
            self.assertEqual(
                int(psm_id_output_new[i][GROUP_ID_INDEX]),
                int(psm_id_output[i][GROUP_ID_INDEX]),
            )
            self.assertEqual(
                set(psm_id_output_new[i][PEPTIDES_INDEX:]),
                set(psm_id_output[i][PEPTIDES_INDEX:]),
            )

        pipeline = pyproteininference.pipeline.ProteinInferencePipeline(
            parameter_file=PARAMETER_FILE,
            database_file=TEST_DATABASE,
            target_files=TARGET_FILE,
            decoy_files=DECOY_FILE,
            combined_files=None,
            output_directory=os.path.join(OUTPUT_DIR, "leads"),
            id_splitting=True,
            append_alt_from_db=False,
        )

        pipeline.execute()

        pipeline_protein_groups = pipeline.data.protein_group_objects

        for i in range(len(pipeline_protein_groups)):
            lead_protein_pipeline = pipeline_protein_groups[i].proteins[0]
            lead_protein = protein_groups[i].proteins[0]
            self.assertEqual(lead_protein_pipeline.identifier, lead_protein.identifier)
            self.assertAlmostEqual(lead_protein_pipeline.score, lead_protein.score)
            self.assertEqual(pipeline_protein_groups[i].q_value, protein_groups[i].q_value)
            self.assertEqual(pipeline_protein_groups[i].number_id, protein_groups[i].number_id)
            self.assertEqual(lead_protein_pipeline.peptides, lead_protein.peptides)



import re


class Protein(object):
    """
    The following class is a representation of a Protein that stores characteristics/attributes of a protein for the
        entire analysis.
    We use __slots__ to predefine the attributes the Protein Object can have.
    This is done to speed up runtime of the PI algorithm.

    Attributes:
        identifier (str): String identifier for the Protein object.
        score (float): Float that represents the protein score as output from
            [Score object][pyproteininference.scoring.Score] methods.
        psms (list): List of [Psm][pyproteininference.physical.Psm] objects.
        group_identification (set): Set of group Identifiers that the protein belongs to (int).
        reviewed (bool): True/False on if the identifier is reviewed.
        unreviewed (bool): True/False on if the identifier is reviewed.
        peptides (list): List of non flanking peptide sequences.
        peptide_scores (list): List of Psm scores associated with the protein.
        picked (bool): True/False if the protein passes the picker algo. True if passes. False if does not pass.
        num_peptides (int): Number of peptides that map to the given Protein.
        unique_peptides (list): List of peptide strings that are unique to this protein across the analysis.
        num_unique_peptides (int): Number of unique peptides.
        raw_peptides (list): List of raw peptides. Includes flanking AA and Mods.

    """

    __slots__ = (
        "identifier",
        "score",
        "psms",
        "group_identification",
        "reviewed",
        "unreviewed",
        "peptides",
        "peptide_scores",
        "picked",
        "num_peptides",
        "unique_peptides",
        "num_unique_peptides",
        "raw_peptides",
    )

    def __init__(self, identifier):
        """
        Initialization method for Protein object.

        Args:
            identifier (str): String identifier for the Protein object.

        Example:
            >>> protein = pyproteininference.physical.Protein(identifier = "PRKDC_HUMAN|P78527")

        """
        self.identifier = identifier
        self.score = None
        self.psms = []  # List of psm objects
        self.group_identification = set()
        self.reviewed = False
        self.unreviewed = False
        self.peptides = None  # Sequence info without flanking
        self.peptide_scores = None  # remove
        self.picked = True
        self.num_peptides = None  # remove
        self.unique_peptides = None  # remove
        self.num_unique_peptides = None  # remove
        self.raw_peptides = set()  # Includes Flanking Seq Info

    def get_psm_scores(self):
        """
        Retrieves psm scores for a given protein.

        Returns:
            list: List of psm scores for the given protein.

        """
        score_list = [x.main_score for x in self.psms]
        return score_list

    def get_psm_identifiers(self):
        """
        Retrieves a list of Psm identifiers.

         Returns:
             list: List of Psm identifiers.

        """
        psms = [x.identifier for x in self.psms]
        return psms

    def get_stripped_psm_identifiers(self):
        """
        Retrieves a list of Psm identifiers that have had mods removed and flanking AAs removed.

         Returns:
             list: List of Psm identifiers that have no mods or flanking AAs.

        """
        psms = [x.stripped_peptide for x in self.psms]
        return psms

    def get_unique_peptide_identifiers(self):
        """
        Retrieves the unique set of peptides for a protein.

         Returns:
             set: Set of peptide strings.

        """
        unique_peptides = set(self.get_psm_identifiers())
        return unique_peptides

    def get_unique_stripped_peptide_identifiers(self):
        """
        Retrieves the unique set of peptides for a protein that are stripped.

         Returns:
             set: Set of peptide strings that are stripped of mods and flanking AAs.

        """
        stripped_peptide_identifiers = set(self.get_stripped_psm_identifiers())
        return stripped_peptide_identifiers

    def get_num_psms(self):
        """
        Retrieves the number of Psms.

         Returns:
             int: Number of Psms.

        """
        num_psms = len(self.get_psm_identifiers())
        return num_psms

    def get_num_peptides(self):
        """
        Retrieves the number of peptides.

         Returns:
             int: Number of peptides.

        """
        num_peptides = len(self.get_unique_peptide_identifiers())
        return num_peptides

    def get_psm_ids(self):
        """
        Retrieves the Psm Ids.

         Returns:
            list: List of Psm Ids.

        """
        psm_ids = [x.psm_id for x in self.psms]
        return psm_ids


class Psm(object):
    """
    The following class is a physical Psm class that stores characteristics of a psm for the entire analysis.
    We use __slots__ to predefine the attributes the Psm Object can have.
    This is done to speed up runtime of the PI algorithm.

    Attributes:
        identifier (str): Peptide Identifier: IE "K.DLIDEGH#AATQLVNQLHDVVVENNLSDK.Q".
        percscore (float): Percolator Score from input file if it exists.
        qvalue (float): Q value from input file if it exists.
        pepvalue (float): Pep value from input file if it exists.
        possible_proteins (list): List of protein strings that the Psm maps to based on the digest.
        psm_id (str): String that represents a global identifier for the Psm. Should come from input files.
        custom_score (float): Score that comes from a custom column in the input files.
        main_score (float): The Psm score to be used as the scoring variable for protein scoring. can be
            percscore,qvalue,pepvalue, or custom_score.
        stripped_peptide (str): This is the identifier attribute that has had mods removed and flanking AAs
            removed IE: DLIDEGHAATQLVNQLHDVVVENNLSDK.
        non_flanking_peptide (str): This is the identifier attribute that has had flanking AAs
            removed IE: DLIDEGH#AATQLVNQLHDVVVENNLSDK. #NOTE Mods are still present here.

    """

    __slots__ = (
        "identifier",
        "percscore",
        "qvalue",
        "pepvalue",
        "possible_proteins",
        "psm_id",
        "custom_score",
        "main_score",
        "stripped_peptide",
        "non_flanking_peptide",
    )

    # The regex removes anything between parantheses including parenthases - \([^()]*\)
    # The regex removes anything between brackets including parenthases - \[.*?\]
    # And the regex removes anything that is not an A-Z character [^A-Z]
    MOD_REGEX = re.compile("\([^()]*\)|\[.*?\]|[^A-Z]")  # noqa W605

    FRONT_FLANKING_REGEX = re.compile("^[A-Z|-][.]")
    BACK_FLANKING_REGEX = re.compile("[.][A-Z|-]$")

    SCORE_ATTRIBUTE_NAMES = set(["pepvalue", "qvalue", "percscore", "custom_score"])

    def __init__(self, identifier):
        """
        Initialization method for the Psm object.
        This method also initializes the `stripped_peptide` and `non_flanking_peptide` attributes.

        Args:
            identifier (str): Peptide Identifier: IE ""K.DLIDEGH#AATQLVNQLHDVVVENNLSDK.Q".

        Example:
            >>> psm = pyproteininference.physical.Psm(identifier = "K.DLIDEGHAATQLVNQLHDVVVENNLSDK.Q")

        """
        self.identifier = identifier
        self.percscore = None
        self.qvalue = None
        self.pepvalue = None
        self.possible_proteins = None
        self.psm_id = None
        self.custom_score = None
        self.main_score = None
        self.stripped_peptide = None
        self.non_flanking_peptide = None

        # Add logic to split the peptide and strip it of mods
        current_peptide = Psm.split_peptide(peptide_string=self.identifier)

        self.non_flanking_peptide = current_peptide

        if not current_peptide.isupper() or not current_peptide.isalpha():
            # If we have mods remove them...
            peptide_string = current_peptide.upper()
            stripped_peptide = Psm.remove_peptide_mods(peptide_string)
            current_peptide = stripped_peptide

        # Set stripped_peptide variable
        self.stripped_peptide = current_peptide

    @classmethod
    def remove_peptide_mods(cls, peptide_string):
        """
        This class method takes a string and uses a `MOD_REGEX` to remove mods from peptide strings.

        Args:
            peptide_string (str): Peptide string to have mods removed from.

        Returns:
            str: a peptide string with mods removed.

        """
        stripped_peptide = cls.MOD_REGEX.sub("", peptide_string)
        return stripped_peptide

    @classmethod
    def split_peptide(cls, peptide_string, delimiter="."):
        """
        This class method takes a peptide string with flanking AAs and removes them from the peptide string.
        This method uses string splitting and if the method produces a faulty peptide the method
            [split_peptide_pro][pyproteininference.physical.Psm.split_peptide_pro] will be called.

        Args:
            peptide_string (str): Peptide string to have mods removed from.
            delimiter (str): a string to indicate what separates a leading/trailing (flanking) AA from the
                peptide sequence.

        Returns:
            str: a peptide string with flanking AAs removed.

        """
        peptide_split = peptide_string.split(delimiter)
        if len(peptide_split) == 3:
            # If we get 3 chunks it will usually be ['A', 'ADGSDFGSS', 'F']
            # So take index 1
            peptide = peptide_split[1]
        elif len(peptide_split) == 1:
            # If we get 1 chunk it should just be ['ADGSDFGSS']
            # So take index 0
            peptide = peptide_split[0]
        else:
            # If we split the peptide and it is not length 1 or 3 then try to split with pro
            peptide = cls.split_peptide_pro(peptide_string=peptide_string, delimiter=delimiter)

        return peptide

    @classmethod
    def split_peptide_pro(cls, peptide_string, delimiter="."):
        """
        This class method takes a peptide string with flanking AAs and removes them from the peptide string.
        This is a specialized method of [split_peptide][pyproteininference.physical.Psm.split_peptide] that uses
         regex identifiers to replace flanking AAs as opposed to string splitting.


        Args:
            peptide_string (str): Peptide string to have mods removed from.
            delimiter (str): a string to indicate what separates a leading/trailing (flanking) AA from the peptide
                sequence.

        Returns:
            str: a peptide string with flanking AAs removed.

        """

        if delimiter != ".":
            front_regex = "^[A-Z|-][{}]".format(delimiter)
            cls.FRONT_FLANKING_REGEX = re.compile(front_regex)
            back_regex = "[{}][A-Z|-]$".format(delimiter)
            cls.BACK_FLANKING_REGEX = re.compile(back_regex)

        # Replace the front flanking with nothing
        peptide_string = cls.FRONT_FLANKING_REGEX.sub("", peptide_string)

        # Replace the back flanking with nothing
        peptide_string = cls.BACK_FLANKING_REGEX.sub("", peptide_string)

        return peptide_string

    def assign_main_score(self, score):
        """
        This method takes in a score type and assigns the variable main_score for a given Psm based on the score type.

        Args:
            score (str): This is a string representation of the Psm attribute that will get assigned to the main_score
                variable.

        """
        # Assign a main score based on user input
        if score not in self.SCORE_ATTRIBUTE_NAMES:
            raise ValueError("Scores must either be one of: '{}'".format(", ".join(self.SCORE_ATTRIBUTE_NAMES)))
        else:
            self.main_score = getattr(self, score)


class ProteinGroup(object):
    """
    The following class is a physical Protein Group class that stores characteristics of a Protein Group for the entire
        analysis.
    We use __slots__ to predefine the attributes the Psm Object can have.
    This is done to speed up runtime of the PI algorithm.

    Attributes:
        number_id (int): unique Integer to represent a group.
        proteins (list): List of [Protein][pyproteininference.physical.Protein] objects.
        q_value (float): Q value for the protein group that is calculated with method
            [calculate_q_values][pyproteininference.datastore.DataStore.calculate_q_values].

    """

    __slots__ = ("proteins", "number_id", "q_value")

    def __init__(self, number_id):
        """
        Initialization method for ProteinGroup object.

        Args:
            number_id (int): unique Integer to represent a group.

        Example:
            >>> pg = pyproteininference.physical.ProteinGroup(number_id = 1)
        """

        self.proteins = []
        self.number_id = number_id
        self.q_value = None



import csv
import os
import tempfile
from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference
from pyproteininference import in_silico_digest
from pyproteininference.parameters import ProteinInferenceParameter

TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_inclusion.yaml")
OUTPUT_DIR = tempfile.gettempdir()
# OUTPUT_DIR = resource_filename('pyproteininference', '../tests/output/')
for sub_dir in ["leads", "all", "peptides", "psms", "psm_ids"]:
    if not os.path.exists(os.path.join(OUTPUT_DIR, sub_dir)):
        os.makedirs(os.path.join(OUTPUT_DIR, sub_dir))

LEAD_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/leads/test_inclusion_q_value_leads_ml_posterior_error_prob.csv",
)
ALL_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/all/test_inclusion_q_value_all_ml_posterior_error_prob.csv",
)
PEPTIDE_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/peptides/test_inclusion_q_value_leads_peptides_ml_posterior_error_prob.csv",
)
PSM_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/psms/test_inclusion_q_value_leads_psms_ml_posterior_error_prob.csv",
)
PSM_ID_OUTPUT_FILE = resource_filename(
    "pyproteininference",
    "../tests/output/psm_ids/test_inclusion_q_value_leads_psm_ids_ml_posterior_error_prob.csv",
)

IDENTIFIER_INDEX = 0
SCORE_INDEX = 1
Q_VALUE_INDEX = 2
GROUP_ID_INDEX = 5
PEPTIDES_INDEX = 6


class TestMissingDatabasePipeline(TestCase):
    def test_missing_database_pipeline(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = in_silico_digest.PyteomicsDigest(
            database_path=None,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=False,
        )
        pep_and_prot_data.read_psms()

        # Make sure we still have 27 psms even without a database digest
        self.assertEqual(len(pep_and_prot_data.psms), 27)

        self.assertDictEqual(
            digest.peptide_to_protein_dictionary,
            {
                "CQTCGYKFHEHCSTK": {"RAF1_HUMAN|P04049"},
                "WHGDVAVKILK": {"RAF1_HUMAN|P04049"},
                "QTAQGMDYLHAK": {
                    "ARAF_HUMAN|P10398",
                    "BRAF_HUMAN|P15056",
                    "RAF1_HUMAN|P04049",
                },
                "SASEPSLHR": {"ARAF_HUMAN|P10398", "RAF1_HUMAN|P04049"},
                "VFLPNKQR": {"BRAF_HUMAN|P15056", "RAF1_HUMAN|P04049"},
                "GYLSPDLSK": {"ARAF_HUMAN|P10398", "BRAF_HUMAN|P15056"},
                "TFFSLAFCDFCLK": {"ARAF_HUMAN|P10398"},
                "FQMFQLIDIAR": {"RAF1_HUMAN|P04049"},
                "LYLLTQMPH": {"TCAF1_HUMAN|Q9Y4C2"},
                "YCWMSTGLYIPGR": {"TCAF1_HUMAN|Q9Y4C2"},
                "AEGGGGGGRPGAPAAGDGK": {"HNRPU_HUMAN|Q00839", "B3KX72_HUMAN|B3KX72"},
                "LQAALDDEEAGGRPAMEPGNGSLDLGGDSAGR": {
                    "Q96BA7_HUMAN|Q96BA7",
                    "HNRPU_HUMAN|Q00839",
                    "B3KX72_HUMAN|B3KX72",
                },
                "CGVEVTQTK": {"RPOC_SHIF8|Q0SY12"},
                "IALASPDMIR": {"RPOC_SHIF8|Q0SY12"},
                "MGAEAIQALLK": {"RPOC_SHIF8|Q0SY12"},
                "RVDYSGR": {"RPOC_SHIF8|Q0SY12"},
                "VIDIWAAANDR": {"RPOC_SHIF8|Q0SY12"},
                "EGLNVLQYFISTHGAR": {"RPOC_SHIF8|Q0SY12"},
                "FATSDLNDLYR": {"RPOC_SHIF8|Q0SY12"},
                "IPQESGGTK": {"RPOC_SHIF8|Q0SY12"},
                "LIPAGTGYAYHQDR": {"RPOC_SHIF8|Q0SY12"},
                "NTLLHEQWCDLLEENSVDAVK": {"RPOC_SHIF8|Q0SY12"},
                "VADLFEAR": {"RPOC_SHIF8|Q0SY12"},
                "VTAEDVLKPGTADILVPR": {"RPOC_SHIF8|Q0SY12"},
                "HPMQTLLYLK": {"##TCAF1_HUMAN|##Q9Y4C2"},
                "MEPTPVPFCGAK": {"##TCAF2_HUMAN|##A6NFQ2"},
                "TVVNVR": {"RAF1_HUMAN|P04049"},
            },
        )

        self.assertDictEqual(
            digest.protein_to_peptide_dictionary,
            {
                "RAF1_HUMAN|P04049": {
                    "CQTCGYKFHEHCSTK",
                    "FQMFQLIDIAR",
                    "SASEPSLHR",
                    "VFLPNKQR",
                    "TVVNVR",
                    "WHGDVAVKILK",
                    "QTAQGMDYLHAK",
                },
                "ARAF_HUMAN|P10398": {
                    "GYLSPDLSK",
                    "SASEPSLHR",
                    "TFFSLAFCDFCLK",
                    "QTAQGMDYLHAK",
                },
                "BRAF_HUMAN|P15056": {"GYLSPDLSK", "VFLPNKQR", "QTAQGMDYLHAK"},
                "TCAF1_HUMAN|Q9Y4C2": {"LYLLTQMPH", "YCWMSTGLYIPGR"},
                "HNRPU_HUMAN|Q00839": {
                    "LQAALDDEEAGGRPAMEPGNGSLDLGGDSAGR",
                    "AEGGGGGGRPGAPAAGDGK",
                },
                "B3KX72_HUMAN|B3KX72": {
                    "LQAALDDEEAGGRPAMEPGNGSLDLGGDSAGR",
                    "AEGGGGGGRPGAPAAGDGK",
                },
                "Q96BA7_HUMAN|Q96BA7": {"LQAALDDEEAGGRPAMEPGNGSLDLGGDSAGR"},
                "RPOC_SHIF8|Q0SY12": {
                    "RVDYSGR",
                    "VIDIWAAANDR",
                    "IPQESGGTK",
                    "EGLNVLQYFISTHGAR",
                    "LIPAGTGYAYHQDR",
                    "IALASPDMIR",
                    "FATSDLNDLYR",
                    "VADLFEAR",
                    "VTAEDVLKPGTADILVPR",
                    "CGVEVTQTK",
                    "NTLLHEQWCDLLEENSVDAVK",
                    "MGAEAIQALLK",
                },
                "##TCAF1_HUMAN|##Q9Y4C2": {"HPMQTLLYLK"},
                "##TCAF2_HUMAN|##A6NFQ2": {"MEPTPVPFCGAK"},
            },
        )

        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        # Make sure we have data in the data dictionaries.
        self.assertEqual(len(data.peptide_to_protein_dictionary().keys()), 27)
        self.assertEqual(len(data.protein_to_peptide_dictionary().keys()), 10)

        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        data.restrict_psm_data()

        self.assertEqual(len(data.main_data_restricted), 26)

        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        data.create_scoring_input()

        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        if protein_inference_parameters.inference_type == pyproteininference.inference.Inference.EXCLUSION:
            # This gets ran if we run exclusion...
            data.exclude_non_distinguishing_peptides()

        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        score = pyproteininference.scoring.Score(data=data)
        score.score_psms(score_method=protein_inference_parameters.protein_score)

        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        if protein_inference_parameters.picker:
            data.protein_picker()
        else:
            pass

        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        inference_type = protein_inference_parameters.inference_type

        if inference_type == pyproteininference.inference.Inference.PARSIMONY:
            group = pyproteininference.inference.Parsimony(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.INCLUSION:
            group = pyproteininference.inference.Inclusion(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.EXCLUSION:
            group = pyproteininference.inference.Exclusion(data=data, digest=digest)
            group.infer_proteins()

        # STEP 11: Run FDR and Q value Calculations
        # STEP 11: Run FDR and Q value Calculations
        # STEP 11: Run FDR and Q value Calculations
        data.calculate_q_values()

        # STEP 12: Export to CSV
        # STEP 12: Export to CSV
        # STEP 12: Export to CSV
        export_type = protein_inference_parameters.export
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=os.path.join(OUTPUT_DIR, "leads"), export_type=export_type)

        lead_output = []
        with open(LEAD_OUTPUT_FILE, "r") as lead_output_file:
            reader = csv.reader(lead_output_file, delimiter=",")
            for row in reader:
                lead_output.append(row)

        del lead_output[0]

        protein_groups = data.protein_group_objects

        for i in range(len(protein_groups)):
            lead_protein = protein_groups[i].proteins[0]
            self.assertEqual(lead_protein.identifier, lead_output[i][IDENTIFIER_INDEX])
            self.assertAlmostEqual(lead_protein.score, float(lead_output[i][SCORE_INDEX]))
            self.assertEqual(protein_groups[i].q_value, float(lead_output[i][Q_VALUE_INDEX]))
            self.assertEqual(protein_groups[i].number_id, int(lead_output[i][GROUP_ID_INDEX]))
            self.assertEqual(lead_protein.peptides, set(lead_output[i][PEPTIDES_INDEX:]))

        export.export_to_csv(directory=os.path.join(OUTPUT_DIR, "all"), export_type="q_value_all")

        all_output = []
        with open(ALL_OUTPUT_FILE, "r") as all_output_file:
            reader = csv.reader(all_output_file, delimiter=",")
            for row in reader:
                all_output.append(row)

        del all_output[0]

        all_output_new = []
        with open(export.filepath, "r") as all_output_file_new:
            reader = csv.reader(all_output_file_new, delimiter=",")
            for row in reader:
                all_output_new.append(row)

        del all_output_new[0]

        for i in range(len(all_output)):
            self.assertEqual(all_output_new[i][IDENTIFIER_INDEX], all_output[i][IDENTIFIER_INDEX])
            self.assertAlmostEqual(float(all_output_new[i][SCORE_INDEX]), float(all_output[i][SCORE_INDEX]))
            self.assertEqual(
                float(all_output_new[i][Q_VALUE_INDEX]),
                float(all_output[i][Q_VALUE_INDEX]),
            )
            self.assertEqual(
                int(all_output_new[i][GROUP_ID_INDEX]),
                int(all_output[i][GROUP_ID_INDEX]),
            )
            self.assertEqual(
                set(all_output_new[i][PEPTIDES_INDEX:]),
                set(all_output[i][PEPTIDES_INDEX:]),
            )

        export.export_to_csv(directory=os.path.join(OUTPUT_DIR, "peptides"), export_type="peptides")

        peptide_output = []
        with open(PEPTIDE_OUTPUT_FILE, "r") as peptide_output_file:
            reader = csv.reader(peptide_output_file, delimiter=",")
            for row in reader:
                peptide_output.append(row)

        del peptide_output[0]

        peptide_output_new = []
        with open(export.filepath, "r") as peptide_output_file_new:
            reader = csv.reader(peptide_output_file_new, delimiter=",")
            for row in reader:
                peptide_output_new.append(row)

        del peptide_output_new[0]

        for i in range(len(peptide_output)):
            self.assertEqual(
                peptide_output_new[i][IDENTIFIER_INDEX],
                peptide_output[i][IDENTIFIER_INDEX],
            )
            self.assertAlmostEqual(
                float(peptide_output_new[i][SCORE_INDEX]),
                float(peptide_output[i][SCORE_INDEX]),
            )
            self.assertEqual(
                float(peptide_output_new[i][Q_VALUE_INDEX]),
                float(peptide_output[i][Q_VALUE_INDEX]),
            )
            self.assertEqual(
                int(peptide_output_new[i][GROUP_ID_INDEX]),
                int(peptide_output[i][GROUP_ID_INDEX]),
            )
            self.assertEqual(
                set(peptide_output_new[i][PEPTIDES_INDEX:]),
                set(peptide_output[i][PEPTIDES_INDEX:]),
            )

        export.export_to_csv(directory=os.path.join(OUTPUT_DIR, "psms"), export_type="psms")

        psm_output = []
        with open(PSM_OUTPUT_FILE, "r") as psm_output_file:
            reader = csv.reader(psm_output_file, delimiter=",")
            for row in reader:
                psm_output.append(row)

        del psm_output[0]

        psm_output_new = []
        with open(export.filepath, "r") as psm_output_file_new:
            reader = csv.reader(psm_output_file_new, delimiter=",")
            for row in reader:
                psm_output_new.append(row)

        del psm_output_new[0]

        for i in range(len(psm_output)):
            self.assertEqual(psm_output_new[i][IDENTIFIER_INDEX], psm_output[i][IDENTIFIER_INDEX])
            self.assertAlmostEqual(float(psm_output_new[i][SCORE_INDEX]), float(psm_output[i][SCORE_INDEX]))
            self.assertEqual(
                float(psm_output_new[i][Q_VALUE_INDEX]),
                float(psm_output[i][Q_VALUE_INDEX]),
            )
            self.assertEqual(
                int(psm_output_new[i][GROUP_ID_INDEX]),
                int(psm_output[i][GROUP_ID_INDEX]),
            )
            self.assertEqual(
                set(psm_output_new[i][PEPTIDES_INDEX:]),
                set(psm_output[i][PEPTIDES_INDEX:]),
            )

        export.export_to_csv(directory=os.path.join(OUTPUT_DIR, "psm_ids"), export_type="psm_ids")

        psm_id_output = []
        with open(PSM_ID_OUTPUT_FILE, "r") as psm_id_output_file:
            reader = csv.reader(psm_id_output_file, delimiter=",")
            for row in reader:
                psm_id_output.append(row)

        del psm_id_output[0]

        psm_id_output_new = []
        with open(export.filepath, "r") as psm_id_output_file_new:
            reader = csv.reader(psm_id_output_file_new, delimiter=",")
            for row in reader:
                psm_id_output_new.append(row)

        del psm_id_output_new[0]

        for i in range(len(psm_id_output)):
            self.assertEqual(
                psm_id_output_new[i][IDENTIFIER_INDEX],
                psm_id_output[i][IDENTIFIER_INDEX],
            )
            self.assertAlmostEqual(
                float(psm_id_output_new[i][SCORE_INDEX]),
                float(psm_id_output[i][SCORE_INDEX]),
            )
            self.assertEqual(
                float(psm_id_output_new[i][Q_VALUE_INDEX]),
                float(psm_id_output[i][Q_VALUE_INDEX]),
            )
            self.assertEqual(
                int(psm_id_output_new[i][GROUP_ID_INDEX]),
                int(psm_id_output[i][GROUP_ID_INDEX]),
            )
            self.assertEqual(
                set(psm_id_output_new[i][PEPTIDES_INDEX:]),
                set(psm_id_output[i][PEPTIDES_INDEX:]),
            )

        pipeline = pyproteininference.pipeline.ProteinInferencePipeline(
            parameter_file=PARAMETER_FILE,
            database_file=None,
            target_files=TARGET_FILE,
            decoy_files=DECOY_FILE,
            combined_files=None,
            output_directory=os.path.join(OUTPUT_DIR, "leads"),
            id_splitting=True,
        )

        pipeline.execute()

        pipeline_protein_groups = pipeline.data.protein_group_objects

        for i in range(len(pipeline_protein_groups)):
            lead_protein_pipeline = pipeline_protein_groups[i].proteins[0]
            lead_protein = protein_groups[i].proteins[0]
            self.assertEqual(lead_protein_pipeline.identifier, lead_protein.identifier)
            self.assertAlmostEqual(lead_protein_pipeline.score, lead_protein.score)
            self.assertEqual(pipeline_protein_groups[i].q_value, protein_groups[i].q_value)
            self.assertEqual(pipeline_protein_groups[i].number_id, protein_groups[i].number_id)
            self.assertEqual(lead_protein_pipeline.peptides, lead_protein.peptides)



from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference
from pyproteininference import in_silico_digest
from pyproteininference.parameters import ProteinInferenceParameter

TEST_DATABASE = resource_filename("pyproteininference", "../tests/data/test_database.fasta")
TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target_no_alt_prot.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy_no_alt_prot.txt")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_inclusion.yaml")

IDENTIFIER_INDEX = 0
SCORE_INDEX = 1
Q_VALUE_INDEX = 2
GROUP_ID_INDEX = 5
PEPTIDES_INDEX = 6


class TestAltProteinDbRecovery(TestCase):
    def test_alt_protein_db_recovery(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data_no_append = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=False,
        )
        pep_and_prot_data_no_append.read_psms()

        self.assertListEqual(
            [x.possible_proteins for x in pep_and_prot_data_no_append.psms],
            [
                ["RAF1_HUMAN|P04049"],
                ["RAF1_HUMAN|P04049"],
                ["ARAF_HUMAN|P10398"],
                ["ARAF_HUMAN|P10398"],
                ["BRAF_HUMAN|P15056"],
                ["ARAF_HUMAN|P10398"],
                ["ARAF_HUMAN|P10398"],
                ["RAF1_HUMAN|P04049"],
                ["TCAF1_HUMAN|Q9Y4C2"],
                ["TCAF1_HUMAN|Q9Y4C2"],
                ["HNRPU_HUMAN|Q00839"],
                ["HNRPU_HUMAN|Q00839"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["##TCAF1_HUMAN|##Q9Y4C2"],
                ["##TCAF2_HUMAN|##A6NFQ2"],
                ["RAF1_HUMAN|P04049"],
            ],
        )

        pep_and_prot_data_append = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=True,
        )
        pep_and_prot_data_append.read_psms()

        self.assertListEqual(
            [x.possible_proteins for x in pep_and_prot_data_append.psms],
            [
                ["RAF1_HUMAN|P04049"],
                ["RAF1_HUMAN|P04049"],
                ["ARAF_HUMAN|P10398", "BRAF_HUMAN|P15056", "RAF1_HUMAN|P04049"],
                ["ARAF_HUMAN|P10398", "RAF1_HUMAN|P04049"],
                ["BRAF_HUMAN|P15056", "RAF1_HUMAN|P04049"],
                ["ARAF_HUMAN|P10398", "BRAF_HUMAN|P15056"],
                ["ARAF_HUMAN|P10398"],
                ["RAF1_HUMAN|P04049"],
                ["TCAF1_HUMAN|Q9Y4C2"],
                ["TCAF1_HUMAN|Q9Y4C2", "TCAF1_HUMAN|Q9Y4C2-2"],
                ["HNRPU_HUMAN|Q00839", "B3KX72_HUMAN|B3KX72", "Q96BA7_HUMAN|Q96BA7"],
                ["HNRPU_HUMAN|Q00839"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["RPOC_SHIF8|Q0SY12"],
                ["##TCAF1_HUMAN|##Q9Y4C2"],
                [
                    "##TCAF2_HUMAN|##A6NFQ2",
                    "##TCAF2_HUMAN|##A6NFQ2-2",
                    "##TCAF2_HUMAN|##A6NFQ2-3",
                ],
                ["RAF1_HUMAN|P04049"],
            ],
        )

        self.assertEqual(len(pep_and_prot_data_append.psms), 27)



import tempfile
from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference

TEST_DATABASE = resource_filename("pyproteininference", "../tests/data/test_database.fasta")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_inclusion.yaml")
OUTPUT_DIR = tempfile.gettempdir()
# OUTPUT_DIR = resource_filename('pyproteininference', '../tests/output/')

TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")

temp_dir = tempfile.gettempdir()


class TestDataStoreMethods(TestCase):
    def test_datastore(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = pyproteininference.parameters.ProteinInferenceParameter(
            yaml_param_filepath=PARAMETER_FILE
        )

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = pyproteininference.in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            combined_files=None,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=False,
        )
        pep_and_prot_data.read_psms()

        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        # STEP 4: Initiate the datastore class #
        data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        # Step 5: Restrict the PSM data
        data.restrict_psm_data()

        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        # Step 6: Generate protein scoring input
        data.create_scoring_input()

        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        # Step 7: Remove non unique peptides if running exclusion
        if protein_inference_parameters.inference_type == pyproteininference.inference.Inference.EXCLUSION:
            # This gets ran if we run exclusion...
            data.exclude_non_distinguishing_peptides()

        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        # STEP 8: Score our PSMs given a score method
        score = pyproteininference.scoring.Score(data=data)
        score.score_psms(score_method=protein_inference_parameters.protein_score)

        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        # STEP 9: Run protein picker on the data
        if protein_inference_parameters.picker:
            data.protein_picker()
        else:
            pass

        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        # STEP 10: Apply Inference
        inference_type = protein_inference_parameters.inference_type

        if inference_type == pyproteininference.inference.Inference.PARSIMONY:
            group = pyproteininference.inference.Parsimony(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.INCLUSION:
            group = pyproteininference.inference.Inclusion(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.EXCLUSION:
            group = pyproteininference.inference.Exclusion(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.FIRST_PROTEIN:
            group = pyproteininference.inference.FirstProtein(data=data, digest=digest)
            group.infer_proteins()

        if inference_type == pyproteininference.inference.Inference.PEPTIDE_CENTRIC:
            group = pyproteininference.inference.PeptideCentric(data=data, digest=digest)
            group.infer_proteins()

        # STEP 11: Run FDR and Q value Calculations
        # STEP 11: Run FDR and Q value Calculations
        # STEP 11: Run FDR and Q value Calculations
        data.calculate_q_values()

        # Start datastore tests

        scored_identifiers = data.get_sorted_identifiers(scored=True)
        self.assertListEqual(
            scored_identifiers,
            [
                "ARAF_HUMAN|P10398",
                "BRAF_HUMAN|P15056",
                "HNRPU_HUMAN|Q00839",
                "RAF1_HUMAN|P04049",
                "RPOC_SHIF8|Q0SY12",
                "TCAF1_HUMAN|Q9Y4C2",
                "##TCAF2_HUMAN|##A6NFQ2",
                "B3KX72_HUMAN|B3KX72",
                "Q96BA7_HUMAN|Q96BA7",
            ],
        )

        identifiers = data.get_sorted_identifiers(scored=False)
        self.assertListEqual(
            identifiers,
            [
                "ARAF_HUMAN|P10398",
                "BRAF_HUMAN|P15056",
                "HNRPU_HUMAN|Q00839",
                "RAF1_HUMAN|P04049",
                "RPOC_SHIF8|Q0SY12",
                "TCAF1_HUMAN|Q9Y4C2",
                "##TCAF1_HUMAN|##Q9Y4C2",
                "##TCAF2_HUMAN|##A6NFQ2",
                "B3KX72_HUMAN|B3KX72",
                "Q96BA7_HUMAN|Q96BA7",
            ],
        )

        psm_data = data.get_psm_data()
        self.assertEqual(len(psm_data), 26)

        protein_data = data.get_protein_data()
        self.assertEqual(len(protein_data), 9)

        proteins_from_psms = data.get_protein_identifiers_from_psm_data()
        self.assertEqual(len(proteins_from_psms), 26)

        q_values = data.get_q_values()
        self.assertEqual(len(q_values), 26)

        pep_values = data.get_pep_values()
        self.assertEqual(len(pep_values), 26)

        protein_info_dict = data.get_protein_information_dictionary()
        self.assertDictEqual(
            protein_info_dict["RAF1_HUMAN|P04049"][0],
            {
                "peptide": "R.CQTCGYKFHEHCSTK.V",
                "Qvalue": 0.00032,
                "PosteriorErrorProbability": 3.5e-06,
                "Percscore": 7.0,
            },
        )

        # Restrict with all filteres
        protein_inference_parameters.restrict_pep = 0.05
        protein_inference_parameters.restrict_q = 0.001
        protein_inference_parameters.restrict_peptide_length = 9

        data.restrict_psm_data(remove1pep=True)
        self.assertEqual(len(data.main_data_restricted), 8)

        protein_inference_parameters.restrict_pep = None
        # Restrict length and q
        data.restrict_psm_data(remove1pep=True)
        self.assertEqual(len(data.main_data_restricted), 8)

        protein_inference_parameters.restrict_q = None
        # Restrict length only
        data.restrict_psm_data(remove1pep=True)
        self.assertEqual(len(data.main_data_restricted), 23)

        protein_inference_parameters.restrict_peptide_length = None
        # No Restriction
        data.restrict_psm_data(remove1pep=True)
        self.assertEqual(len(data.main_data_restricted), 27)

        # No restriction and no removing 1 pep values
        data.restrict_psm_data(remove1pep=False)
        self.assertEqual(len(data.main_data_restricted), 27)

        protein_inference_parameters.restrict_pep = 0.05
        protein_inference_parameters.restrict_q = 0.05
        # Restrict pep and q without length
        data.restrict_psm_data(remove1pep=True)
        self.assertEqual(len(data.main_data_restricted), 24)

        protein_inference_parameters.restrict_q = 0.001
        protein_inference_parameters.restrict_pep = None
        protein_inference_parameters.restrict_peptide_length = None
        # restrict pep and length without q
        data.restrict_psm_data(remove1pep=True)
        self.assertEqual(len(data.main_data_restricted), 9)

        protein_inference_parameters.restrict_q = None
        protein_inference_parameters.restrict_pep = 0.003
        protein_inference_parameters.restrict_peptide_length = None
        # restrict pep and length without q
        data.restrict_psm_data(remove1pep=True)
        self.assertEqual(len(data.main_data_restricted), 24)

        protein_inference_parameters.restrict_q = None
        protein_inference_parameters.restrict_pep = 0.0004
        protein_inference_parameters.restrict_peptide_length = 9
        # restrict pep and length without q
        data.restrict_psm_data(remove1pep=True)
        self.assertEqual(len(data.main_data_restricted), 10)

        prot_dict = data.protein_to_peptide_dictionary()
        self.assertSetEqual(
            prot_dict["RAF1_HUMAN|P04049"],
            set(
                [
                    "FQMFQLIDIAR",
                    "WHGDVAVKILK",
                    "SASEPSLHR",
                    "CQTCGYKFHEHCSTK",
                    "QTAQGMDYLHAK",
                ]
            ),
        )
        pep_dict = data.peptide_to_protein_dictionary()
        self.assertSetEqual(
            pep_dict["QTAQGMDYLHAK"],
            set(["ARAF_HUMAN|P10398", "RAF1_HUMAN|P04049", "BRAF_HUMAN|P15056"]),
        )

        utl = data.unique_to_leads_peptides()
        self.assertSetEqual(
            utl,
            set(
                [
                    "FQMFQLIDIAR",
                    "TFFSLAFCDFCLK",
                    "WHGDVAVKILK",
                    "CGVEVTQTK",
                    "RVDYSGR",
                    "VTAEDVLKPGTADILVPR",
                    "IALASPDMIR",
                    "CQTCGYKFHEHCSTK",
                    "YCWMSTGLYIPGR",
                    "NTLLHEQWCDLLEENSVDAVK",
                    "VADLFEAR",
                    "VIDIWAAANDR",
                    "MGAEAIQALLK",
                    "IPQESGGTK",
                    "LIPAGTGYAYHQDR",
                    "MEPTPVPFCGAK",
                    "FATSDLNDLYR",
                    "LYLLTQMPH",
                    "EGLNVLQY#FISTHGAR",
                ]
            ),
        )
        hl = data.higher_or_lower()
        self.assertEqual(hl, "higher")

        df_main = data.get_protein_identifiers(data_form="main")
        self.assertListEqual(
            sorted(df_main[2][0]),
            sorted(["ARAF_HUMAN|P10398", "RAF1_HUMAN|P04049", "BRAF_HUMAN|P15056"]),
        )
        df_restricted = data.get_protein_identifiers(data_form="restricted")
        self.assertEqual(len(df_restricted), 10)
        df_picked = data.get_protein_identifiers(data_form="picked")
        self.assertListEqual(
            df_picked,
            [
                "RPOC_SHIF8|Q0SY12",
                "RAF1_HUMAN|P04049",
                "ARAF_HUMAN|P10398",
                "BRAF_HUMAN|P15056",
                "TCAF1_HUMAN|Q9Y4C2",
                "B3KX72_HUMAN|B3KX72",
                "HNRPU_HUMAN|Q00839",
                "Q96BA7_HUMAN|Q96BA7",
                "##TCAF2_HUMAN|##A6NFQ2",
            ],
        )
        df_picked_removed = data.get_protein_identifiers(data_form="picked_removed")
        self.assertListEqual(df_picked_removed, ["##TCAF1_HUMAN|##Q9Y4C2"])

        p1 = data.get_protein_information(protein_string="RAF1_HUMAN|P04049")
        self.assertListEqual(
            p1[1],
            [
                "RAF1_HUMAN|P04049",
                70.7434325345954,
                {2},
                True,
                {
                    "SASEPSLHR",
                    "QTAQGMDYLHAK",
                    "VFLPNKQR",
                    "WHGDVAVKILK",
                    "FQMFQLIDIAR",
                    "CQTCGYKFHEHCSTK",
                },
                None,
                True,
                6,
            ],
        )



import copy
import csv
import logging
import os
import sys

from pyproteininference.datastore import DataStore
from pyproteininference.inference import Inference
from pyproteininference.physical import Psm
from pyproteininference.scoring import Score

logger = logging.getLogger(__name__)

# set up our logger
logging.basicConfig(
    stream=sys.stderr,
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


class Reader(object):
    """
    Main Reader Class which is parent to all reader subclasses.

    Attributes:
        target_file (str/list): Path to Target PSM result files.
        decoy_file (str/list): Path to Decoy PSM result files.
        combined_files (str/list): Path to Combined PSM result files.
        directory (str): Path to directory containing combined PSM result files.

    """

    MAX_ALLOWED_ALTERNATIVE_PROTEINS = 50

    def __init__(self, target_file=None, decoy_file=None, combined_files=None, directory=None):
        """

        Args:
            target_file (str/list): Path to Target PSM result files.
            decoy_file (str/list): Path to Decoy PSM result files.
            combined_files (str/list): Path to Combined PSM result files.
            directory (str): Path to directory containing combined PSM result files.

        """
        self.target_file = target_file
        self.decoy_file = decoy_file
        self.combined_files = combined_files
        self.directory = directory

    def get_alternative_proteins_from_input(self, row):
        """
        Method to get the alternative proteins from the input files.

        """
        if None in row.keys():
            try:
                row["alternative_proteins"] = row.pop(None)
                # Sort the alternative proteins - when they are read in they become unsorted
                row["alternative_proteins"] = sorted(row["alternative_proteins"])
            except KeyError:
                row["alternative_proteins"] = []
        else:
            row["alternative_proteins"] = []
        return row

    def _validate_input(self):
        """
        Internal method to validate the input to Reader.

        """
        if self.target_file and self.decoy_file and not self.combined_files and not self.directory:
            logger.info("Validating input as target_file and decoy_file")
        elif self.combined_files and not self.target_file and not self.decoy_file and not self.directory:
            logger.info("Validating input as combined_files")
        elif self.directory and not self.combined_files and not self.decoy_file and not self.target_file:
            logger.info("Validating input as combined_directory")
        else:
            raise ValueError(
                "To run Protein inference please supply either: "
                "(1) either one or multiple target_files and decoy_files, "
                "(2) either one or multiple combined_files that include target and decoy data"
                "(3) a combined_directory that contains combined target/decoy files (combined_directory)"
            )

    @classmethod
    def _fix_alternative_proteins(
        cls,
        append_alt_from_db,
        identifiers_sorted,
        max_proteins,
        psm,
        parameter_file_object,
    ):
        """
        Internal method to fix the alternative proteins variable for a given
         [Psm][pyproteininference.physical.Psm] object.

        Args:
            append_alt_from_db (bool): Whether or not to append alternative proteins found in the database that are
                not in the input files.
            identifiers_sorted (list): List of sorted Protein Strings for the given Psm.
            max_proteins (int): Maximum number of proteins that a [Psm][pyproteininference.physical.Psm]
                is allowed to map to.
            psm: (Psm): [Psm][pyproteininference.physical.Psm] object of interest.
            parameter_file_object: (ProteinInferenceParameter):
                [ProteinInferenceParameter][pyproteininference.parameters.ProteinInferenceParameter].

        Returns:
            pyproteininference.physical.Psm: [Psm][pyproteininference.physical.Psm] with alternative proteins fixed.

        """
        # If we are appending alternative proteins from the db
        if append_alt_from_db:
            # Loop over the Identifiers from the DB These are identifiers that contain the current peptide
            for alt_proteins in identifiers_sorted[:max_proteins]:
                # If the identifier is not already in possible proteins
                # and if then len of poss prot is less than the max...
                # Then append
                if alt_proteins not in psm.possible_proteins and len(psm.possible_proteins) < max_proteins:
                    psm.possible_proteins.append(alt_proteins)
        # Next if the len of possible proteins is greater than max then restrict the list length...
        if len(psm.possible_proteins) > max_proteins:
            psm.possible_proteins = [psm.possible_proteins[x] for x in range(max_proteins)]
        else:
            pass

        # If no inference only select first poss protein
        if parameter_file_object.inference_type == Inference.FIRST_PROTEIN:
            psm.possible_proteins = [psm.possible_proteins[0]]

        return psm

    def _check_initial_database_overlap(self, initial_possible_proteins, initial_protein_peptide_map):
        """
        Internal method that checks to make sure there is at least some overlap between proteins in the input files
        And the proteins in the database digestion.
        """

        if len(initial_protein_peptide_map.keys()) > 0:
            input_protein_ids_flat = set([protein for group in initial_possible_proteins for protein in group])

            digest_proteins = set(initial_protein_peptide_map.keys())

            intersection = input_protein_ids_flat.intersection(digest_proteins)

            if len(intersection) < 1:
                raise ValueError(
                    "The Intersection of Protein Identifiers between the database digest "
                    "and the input files is zero. Please consider setting id_splitting to True. "
                    "Or make sure that the identifiers in the input files and database file match. "
                    "Example Protein Identifier from input file '{}'."
                    "Example Protein Identifier from database file '{}'".format(
                        list(input_protein_ids_flat)[0], list(digest_proteins)[0]
                    )
                )
            else:
                logger.info("Number of matching proteins from database and input files: {}".format(len(intersection)))
                logger.info("Number of proteins from database file: {}".format(len(digest_proteins)))
                logger.info("Number of proteins from input files: {}".format(len(input_protein_ids_flat)))

        else:
            pass


class PercolatorReader(Reader):
    """
    The following class takes a percolator target file and a percolator decoy file
    or combined files/directory and creates standard [Psm][pyproteininference.physical.Psm] objects.
    This reader class is used as input for [DataStore object][pyproteininference.datastore.DataStore].

    Percolator Output is formatted as follows:
    with each entry being tab delimited.

    | PSMId                         | score    |  q-value    | posterior_error_prob  |  peptide                       | proteinIds          |                      |                      |                         | # noqa E501 W605
    |-------------------------------|----------|-------------|-----------------------|--------------------------------|---------------------|----------------------|----------------------|-------------------------| # noqa E501 W605
    |     116108.15139.15139.6.dta  |  3.44016 | 0.000479928 | 7.60258e-10           | K.MVVSMTLGLHPWIANIDDTQYLAAK.R  | CNDP1_HUMAN\|Q96KN2 | B4E180_HUMAN\|B4E180 | A8K1K1_HUMAN\|A8K1K1 | J3KRP0_HUMAN\|J3KRP0    | # noqa E501 W605

    Attributes:
        target_file (str/list): Path to Target PSM result files.
        decoy_file (str/list): Path to Decoy PSM result files.
        combined_files (str/list): Path to Combined PSM result files.
        directory (str): Path to directory containing combined PSM result files.
        PSMID_INDEX (int): Index of the PSMId from the input files.
        PERC_SCORE_INDEX (int): Index of the Percolator score from the input files.
        Q_VALUE_INDEX (int): Index of the q-value from the input files.
        POSTERIOR_ERROR_PROB_INDEX (int): Index of the posterior error probability from the input files.
        PEPTIDE_INDEX (int): Index of the peptides from the input files.
        PROTEINIDS_INDEX (int): Index of the proteins from the input files.
        psms (list): List of [Psm][pyproteininference.physical.Psm] objects.

    """

    PSMID_INDEX = 0
    PERC_SCORE_INDEX = 1
    Q_VALUE_INDEX = 2
    POSTERIOR_ERROR_PROB_INDEX = 3
    PEPTIDE_INDEX = 4
    PROTEINIDS_INDEX = 5

    def __init__(
        self,
        digest,
        parameter_file_object,
        append_alt_from_db=True,
        target_file=None,
        decoy_file=None,
        combined_files=None,
        directory=None,
    ):
        """

        Args:
            digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].
            parameter_file_object (ProteinInferenceParameter):
                [ProteinInferenceParameter][pyproteininference.parameters.ProteinInferenceParameter].
            append_alt_from_db (bool): Whether or not to append alternative proteins found in the database that
                are not in the input files.
            target_file (str/list): Path to Target PSM result files.
            decoy_file (str/list): Path to Decoy PSM result files.
            combined_files (str/list): Path to Combined PSM result files.
            directory (str): Path to directory containing combined PSM result files.

        Returns:
            Reader: [Reader][pyproteininference.reader.Reader] object.

        Example:
            >>> pyproteininference.reader.PercolatorReader(target_file = "example_target.txt",
            >>>     decoy_file = "example_decoy.txt", digest=digest,parameter_file_object=pi_params)
        """
        self.target_file = target_file
        self.decoy_file = decoy_file
        self.combined_files = combined_files
        self.directory = directory
        # Define Indicies based on input

        self.psms = None
        self.search_id = None
        self.digest = digest
        self.initial_protein_peptide_map = copy.copy(self.digest.protein_to_peptide_dictionary)
        self.append_alt_from_db = append_alt_from_db

        self.parameter_file_object = parameter_file_object

        self._validate_input()

    def read_psms(self):
        """
        Method to read psms from the input files and to transform them into a list of
        [Psm][pyproteininference.physical.Psm] objects.

        This method sets the `psms` variable. Which is a list of Psm objets.

        This method must be ran before initializing [DataStore object][pyproteininference.datastore.DataStore].

        Example:
            >>> reader = pyproteininference.reader.PercolatorReader(target_file = "example_target.txt",
            >>>     decoy_file = "example_decoy.txt",
            >>>     digest=digest, parameter_file_object=pi_params)
            >>> reader.read_psms()

        """
        # Read in and split by line
        if self.target_file and self.decoy_file:
            # If target_file is a list... read them all in and concatenate...
            if isinstance(self.target_file, (list,)):
                all_target = []
                for t_files in self.target_file:
                    logger.info(t_files)
                    ptarg = []
                    with open(t_files, "r") as perc_target_file:
                        spamreader = csv.reader(perc_target_file, delimiter="\t")
                        for row in spamreader:
                            ptarg.append(row)
                    del ptarg[0]
                    all_target = all_target + ptarg
            elif self.target_file:
                # If not just read the file...
                ptarg = []
                with open(self.target_file, "r") as perc_target_file:
                    spamreader = csv.reader(perc_target_file, delimiter="\t")
                    for row in spamreader:
                        ptarg.append(row)
                del ptarg[0]
                all_target = ptarg

            # Repeat for decoy file
            if isinstance(self.decoy_file, (list,)):
                all_decoy = []
                for d_files in self.decoy_file:
                    logger.info(d_files)
                    pdec = []
                    with open(d_files, "r") as perc_decoy_file:
                        spamreader = csv.reader(perc_decoy_file, delimiter="\t")
                        for row in spamreader:
                            pdec.append(row)
                    del pdec[0]
                    all_decoy = all_decoy + pdec
            elif self.decoy_file:
                pdec = []
                with open(self.decoy_file, "r") as perc_decoy_file:
                    spamreader = csv.reader(perc_decoy_file, delimiter="\t")
                    for row in spamreader:
                        pdec.append(row)
                del pdec[0]
                all_decoy = pdec

            # Combine the lists
            perc_all = all_target + all_decoy

        elif self.combined_files:
            if isinstance(self.combined_files, (list,)):
                all = []
                for f in self.combined_files:
                    logger.info(f)
                    combined_psm_result_rows = []
                    with open(f, "r") as perc_files:
                        spamreader = csv.reader(perc_files, delimiter="\t")
                        for row in spamreader:
                            combined_psm_result_rows.append(row)
                    del combined_psm_result_rows[0]
                    all = all + combined_psm_result_rows
            elif self.combined_files:
                # If not just read the file...
                combined_psm_result_rows = []
                with open(self.combined_files, "r") as perc_files:
                    spamreader = csv.reader(perc_files, delimiter="\t")
                    for row in spamreader:
                        combined_psm_result_rows.append(row)
                del combined_psm_result_rows[0]
                all = combined_psm_result_rows
            perc_all = all

        elif self.directory:

            all_files = os.listdir(self.directory)
            all = []
            for files in all_files:
                logger.info(files)
                combined_psm_result_rows = []
                with open(files, "r") as perc_file:
                    spamreader = csv.reader(perc_file, delimiter="\t")
                    for row in spamreader:
                        combined_psm_result_rows.append(row)
                del combined_psm_result_rows[0]
                all = all + combined_psm_result_rows
            perc_all = all

        peptide_to_protein_dictionary = self.digest.peptide_to_protein_dictionary

        perc_all_filtered = []
        for psms in perc_all:
            try:
                float(psms[self.POSTERIOR_ERROR_PROB_INDEX])
                perc_all_filtered.append(psms)
            except ValueError as e:  # noqa F841
                pass

        # Filter by pep
        perc_all = sorted(
            perc_all_filtered,
            key=lambda x: float(x[self.POSTERIOR_ERROR_PROB_INDEX]),
            reverse=False,
        )

        list_of_psm_objects = []
        peptide_tracker = set()
        all_sp_proteins = set(self.digest.swiss_prot_protein_set)
        # We only want to get unique peptides... using all messes up scoring...
        # Create Psm objects with the identifier, percscore, qvalue, pepvalue, and possible proteins...

        initial_poss_prots = []
        logger.info("Length of PSM Data: {}".format(len(perc_all)))
        for psm_info in perc_all:
            current_peptide = psm_info[self.PEPTIDE_INDEX]
            # Define the Psm...
            if current_peptide not in peptide_tracker:
                combined_psm_result_rows = Psm(identifier=current_peptide)
                # Add all the attributes
                combined_psm_result_rows.percscore = float(psm_info[self.PERC_SCORE_INDEX])
                combined_psm_result_rows.qvalue = float(psm_info[self.Q_VALUE_INDEX])
                combined_psm_result_rows.pepvalue = float(psm_info[self.POSTERIOR_ERROR_PROB_INDEX])
                if self.parameter_file_object.inference_type == Inference.FIRST_PROTEIN:
                    poss_proteins = [psm_info[self.PROTEINIDS_INDEX]]
                else:
                    poss_proteins = sorted(list(set(psm_info[self.PROTEINIDS_INDEX :])))  # noqa E203
                    poss_proteins = poss_proteins[: self.MAX_ALLOWED_ALTERNATIVE_PROTEINS]
                combined_psm_result_rows.possible_proteins = poss_proteins  # Restrict to 50 total possible proteins...
                combined_psm_result_rows.psm_id = psm_info[self.PSMID_INDEX]
                input_poss_prots = copy.copy(poss_proteins)

                # Split peptide if flanking
                current_peptide = Psm.split_peptide(peptide_string=current_peptide)

                if not current_peptide.isupper() or not current_peptide.isalpha():
                    # If we have mods remove them...
                    peptide_string = current_peptide.upper()
                    stripped_peptide = Psm.remove_peptide_mods(peptide_string)
                    current_peptide = stripped_peptide

                # Add the other possible_proteins from insilicodigest here...
                try:
                    current_alt_proteins = sorted(
                        list(peptide_to_protein_dictionary[current_peptide])
                    )  # This peptide needs to be scrubbed of Mods...
                except KeyError:
                    current_alt_proteins = []
                    logger.debug(
                        "Peptide {} was not found in the supplied DB with the following proteins {}".format(
                            current_peptide,
                            ";".join(combined_psm_result_rows.possible_proteins),
                        )
                    )
                    for poss_prot in combined_psm_result_rows.possible_proteins:
                        self.digest.peptide_to_protein_dictionary.setdefault(current_peptide, set()).add(poss_prot)
                        self.digest.protein_to_peptide_dictionary.setdefault(poss_prot, set()).add(current_peptide)
                        logger.debug(
                            "Adding Peptide {} and Protein {} to Digest dictionaries".format(current_peptide, poss_prot)
                        )

                # Sort Alt Proteins by Swissprot then Trembl...
                identifiers_sorted = DataStore.sort_protein_strings(
                    protein_string_list=current_alt_proteins,
                    sp_proteins=all_sp_proteins,
                    decoy_symbol=self.parameter_file_object.decoy_symbol,
                )

                # Restrict to 50 possible proteins
                combined_psm_result_rows = self._fix_alternative_proteins(
                    append_alt_from_db=self.append_alt_from_db,
                    identifiers_sorted=identifiers_sorted,
                    max_proteins=self.MAX_ALLOWED_ALTERNATIVE_PROTEINS,
                    psm=combined_psm_result_rows,
                    parameter_file_object=self.parameter_file_object,
                )

                # Remove blank alt proteins
                combined_psm_result_rows.possible_proteins = [
                    x for x in combined_psm_result_rows.possible_proteins if x != ""
                ]

                list_of_psm_objects.append(combined_psm_result_rows)
                peptide_tracker.add(current_peptide)

                initial_poss_prots.append(input_poss_prots)

        self.psms = list_of_psm_objects

        self._check_initial_database_overlap(
            initial_possible_proteins=initial_poss_prots, initial_protein_peptide_map=self.initial_protein_peptide_map
        )

        logger.info("Length of PSM Data: {}".format(len(self.psms)))

        # return perc


class ProteologicPostSearchReader(Reader):
    """
    This class is used to read from post processing proteologic logical object.

    Attributes:
        proteologic_object (list): List of proteologic post search objects.
        search_id (int): Search ID or Search IDs associated with the data.
        postsearch_id (int): PostSearch ID or PostSearch IDs associated with the data.
        digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].
        parameter_file_object (ProteinInferenceParameter):
            [ProteinInferenceParameter][pyproteininference.parameters.ProteinInferenceParameter] object.
        append_alt_from_db (bool): Whether or not to append alternative proteins found in the database
            that are not in the input files.

    """

    def __init__(
        self,
        proteologic_object,
        search_id,
        postsearch_id,
        digest,
        parameter_file_object,
        append_alt_from_db=True,
    ):
        """

        Args:
            proteologic_object (list): List of proteologic post search objects.
            search_id (int): Search ID or Search IDs associated with the data.
            postsearch_id: PostSearch ID or PostSearch IDs associated with the data.
            digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].
            parameter_file_object (ProteinInferenceParameter):
                [ProteinInferenceParameter][pyproteininference.parameters.ProteinInferenceParameter] object.
            append_alt_from_db (bool): Whether or not to append alternative proteins found in the database
                that are not in the input files.


        Returns:
            object:
        """
        self.proteologic_object = proteologic_object
        self.search_id = search_id
        self.postsearch_id = postsearch_id

        self.psms = None
        self.digest = digest
        self.initial_protein_peptide_map = copy.copy(self.digest.protein_to_peptide_dictionary)
        self.append_alt_from_db = append_alt_from_db

        self.parameter_file_object = parameter_file_object

    def read_psms(self):
        """
        Method to read psms from the input files and to transform them into a list of
        [Psm][pyproteininference.physical.Psm] objects.

        This method sets the `psms` variable. Which is a list of Psm objets.

        This method must be ran before initializing [DataStore object][pyproteininference.datastore.DataStore].

        """
        logger.info("Reading in data from Proteologic...")
        if isinstance(self.proteologic_object, (list,)):
            list_of_psms = []
            for p_objs in self.proteologic_object:
                for psms in p_objs.physical_object.psm_sets:
                    list_of_psms.append(psms)
        else:
            list_of_psms = self.proteologic_object.physical_object.psm_sets

        # Sort this by posterior error prob...
        list_of_psms = sorted(list_of_psms, key=lambda x: float(x.psm_filter.pepvalue))

        peptide_to_protein_dictionary = self.digest.peptide_to_protein_dictionary

        list_of_psm_objects = []
        peptide_tracker = set()
        all_sp_proteins = set(self.digest.swiss_prot_protein_set)
        # Peptide tracker is used because we only want UNIQUE peptides...
        # The data is sorted by percolator score... or at least it should be...
        # Or sorted by posterior error probability

        initial_poss_prots = []
        for peps in list_of_psms:
            current_peptide = peps.peptide.sequence
            # Define the Psm...
            if current_peptide not in peptide_tracker:
                p = Psm(identifier=current_peptide)
                # Add all the attributes
                p.percscore = float(0)  # Will be stored in table in future I think...
                p.qvalue = float(peps.psm_filter.q_value)
                p.pepvalue = float(peps.psm_filter.pepvalue)
                if peps.peptide.protein not in peps.alternative_proteins:
                    p.possible_proteins = [peps.peptide.protein] + peps.alternative_proteins
                else:
                    p.possible_proteins = peps.alternative_proteins

                p.possible_proteins = list(filter(None, p.possible_proteins))
                input_poss_prots = copy.copy(p.possible_proteins)
                p.psm_id = peps.spectrum.spectrum_identifier

                # Split peptide if flanking
                current_peptide = Psm.split_peptide(peptide_string=current_peptide)

                if not current_peptide.isupper() or not current_peptide.isalpha():
                    # If we have mods remove them...
                    peptide_string = current_peptide.upper()
                    stripped_peptide = Psm.remove_peptide_mods(peptide_string)
                    current_peptide = stripped_peptide

                # Add the other possible_proteins from insilicodigest here...
                try:
                    current_alt_proteins = sorted(
                        list(peptide_to_protein_dictionary[current_peptide])
                    )  # This peptide needs to be scrubbed of Mods...
                except KeyError:
                    current_alt_proteins = []
                    logger.debug(
                        "Peptide {} was not found in the supplied DB with the following proteins {}".format(
                            current_peptide, ";".join(p.possible_proteins)
                        )
                    )
                    for poss_prot in p.possible_proteins:
                        self.digest.peptide_to_protein_dictionary.setdefault(current_peptide, set()).add(poss_prot)
                        self.digest.protein_to_peptide_dictionary.setdefault(poss_prot, set()).add(current_peptide)
                        logger.debug(
                            "Adding Peptide {} and Protein {} to Digest dictionaries".format(current_peptide, poss_prot)
                        )

                # Sort Alt Proteins by Swissprot then Trembl...
                identifiers_sorted = DataStore.sort_protein_strings(
                    protein_string_list=current_alt_proteins,
                    sp_proteins=all_sp_proteins,
                    decoy_symbol=self.parameter_file_object.decoy_symbol,
                )

                # Restrict to 50 possible proteins... and append alt proteins from db
                p = self._fix_alternative_proteins(
                    append_alt_from_db=self.append_alt_from_db,
                    identifiers_sorted=identifiers_sorted,
                    max_proteins=self.MAX_ALLOWED_ALTERNATIVE_PROTEINS,
                    psm=p,
                    parameter_file_object=self.parameter_file_object,
                )

                list_of_psm_objects.append(p)
                peptide_tracker.add(current_peptide)

                initial_poss_prots.append(input_poss_prots)

        self.psms = list_of_psm_objects

        self._check_initial_database_overlap(
            initial_possible_proteins=initial_poss_prots, initial_protein_peptide_map=self.initial_protein_peptide_map
        )

        logger.info("Finished reading in data from Proteologic...")


class GenericReader(Reader):
    """
    The following class takes a percolator like target file and a percolator like decoy file
    and creates standard [Psm][pyproteininference.physical.Psm] objects.

    Percolator Like Output is formatted as follows:
    with each entry being tab delimited.

    | PSMId                         | score    |  q-value    | posterior_error_prob  |  peptide                       | proteinIds          |                      |                      |                         | # noqa E501 W605
    |-------------------------------|----------|-------------|-----------------------|--------------------------------|---------------------|----------------------|----------------------|-------------------------| # noqa E501 W605
    |     116108.15139.15139.6.dta  |  3.44016 | 0.000479928 | 7.60258e-10           | K.MVVSMTLGLHPWIANIDDTQYLAAK.R  | CNDP1_HUMAN\|Q96KN2 | B4E180_HUMAN\|B4E180 | A8K1K1_HUMAN\|A8K1K1 | J3KRP0_HUMAN\|J3KRP0    | # noqa E501 W605

    Custom columns can be added and used as scoring input. Please see package documentation for more information.

    Attributes:
        target_file (str/list): Path to Target PSM result files.
        decoy_file (str/list): Path to Decoy PSM result files.
        combined_files (str/list): Path to Combined PSM result files.
        directory (str): Path to directory containing combined PSM result files.
        psms (list): List of [Psm][pyproteininference.physical.Psm] objects.
        load_custom_score (bool): True/False on whether or not to load a custom score. Depends on scoring_variable.
        scoring_variable (str): String to indicate which column in the input file is to be used as the scoring input.
        digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].
        parameter_file_object (ProteinInferenceParameter):
            [ProteinInferenceParameter][pyproteininference.parameters.ProteinInferenceParameter] object
        append_alt_from_db (bool): Whether or not to append alternative proteins found in the database that
            are not in the input files.



    """

    PSMID = "PSMId"
    SCORE = "score"
    Q_VALUE = "q-value"
    POSTERIOR_ERROR_PROB = "posterior_error_prob"
    PEPTIDE = "peptide"
    PROTEIN_IDS = "proteinIds"
    ALTERNATIVE_PROTEINS = "alternative_proteins"

    def __init__(
        self,
        digest,
        parameter_file_object,
        append_alt_from_db=True,
        target_file=None,
        decoy_file=None,
        combined_files=None,
        directory=None,
    ):
        """

        Args:
            digest (Digest): [Digest Object][pyproteininference.in_silico_digest.Digest].
            parameter_file_object (ProteinInferenceParameter):
                [ProteinInferenceParameter][pyproteininference.parameters.ProteinInferenceParameter] object.
            append_alt_from_db (bool): Whether or not to append alternative proteins found in the database that
                are not in the input files.
            target_file (str/list): Path to Target PSM result files.
            decoy_file (str/list): Path to Decoy PSM result files.
            combined_files (str/list): Path to Combined PSM result files.
            directory (str): Path to directory containing combined PSM result files.

        Returns:
            Reader: [Reader][pyproteininference.reader.Reader] object.

        Example:
            >>> pyproteininference.reader.GenericReader(target_file = "example_target.txt",
            >>>     decoy_file = "example_decoy.txt",
            >>>     digest=digest, parameter_file_object=pi_params)
        """
        self.target_file = target_file
        self.decoy_file = decoy_file
        self.combined_files = combined_files
        self.directory = directory

        self.psms = None
        self.search_id = None
        self.digest = digest
        self.initial_protein_peptide_map = copy.copy(self.digest.protein_to_peptide_dictionary)
        self.load_custom_score = False

        self.append_alt_from_db = append_alt_from_db

        self.parameter_file_object = parameter_file_object
        self.scoring_variable = parameter_file_object.psm_score

        self._validate_input()

        if self.scoring_variable != self.Q_VALUE and self.scoring_variable != self.POSTERIOR_ERROR_PROB:
            self.load_custom_score = True
            logger.info(
                "Pulling custom column based on parameter file input for score, Column: {}".format(
                    self.scoring_variable
                )
            )
        else:
            logger.info(
                "Pulling no custom columns based on parameter file input for score, using standard Column: {}".format(
                    self.scoring_variable
                )
            )

        # If we select to not run inference at all
        if self.parameter_file_object.inference_type == Inference.FIRST_PROTEIN:
            # Only allow 1 Protein per PSM
            self.MAX_ALLOWED_ALTERNATIVE_PROTEINS = 1

    def read_psms(self):
        """
        Method to read psms from the input files and to transform them into a list of
        [Psm][pyproteininference.physical.Psm] objects.

        This method sets the `psms` variable. Which is a list of Psm objets.

        This method must be ran before initializing [DataStore object][pyproteininference.datastore.DataStore].

        Example:
            >>> reader = pyproteininference.reader.GenericReader(target_file = "example_target.txt",
            >>>     decoy_file = "example_decoy.txt",
            >>>     digest=digest, parameter_file_object=pi_params)
            >>> reader.read_psms()

        """
        logger.info("Reading in Input Files using Generic Reader...")
        # Read in and split by line
        # If target_file is a list... read them all in and concatenate...
        if self.target_file and self.decoy_file:
            if isinstance(self.target_file, (list,)):
                all_target = []
                for t_files in self.target_file:
                    ptarg = []
                    with open(t_files, "r") as psm_target_file:
                        logger.info(t_files)
                        spamreader = csv.DictReader(psm_target_file, delimiter="\t")
                        for row in spamreader:
                            row = self.get_alternative_proteins_from_input(row)
                            ptarg.append(row)
                    all_target = all_target + ptarg
            else:
                # If not just read the file...
                ptarg = []
                with open(self.target_file, "r") as psm_target_file:
                    logger.info(self.target_file)
                    spamreader = csv.DictReader(psm_target_file, delimiter="\t")
                    for row in spamreader:
                        row = self.get_alternative_proteins_from_input(row)
                        ptarg.append(row)
                all_target = ptarg

            # Repeat for decoy file
            if isinstance(self.decoy_file, (list,)):
                all_decoy = []
                for d_files in self.decoy_file:
                    pdec = []
                    with open(d_files, "r") as psm_decoy_file:
                        logger.info(d_files)
                        spamreader = csv.DictReader(psm_decoy_file, delimiter="\t")
                        for row in spamreader:
                            row = self.get_alternative_proteins_from_input(row)
                            pdec.append(row)
                    all_decoy = all_decoy + pdec
            else:
                pdec = []
                with open(self.decoy_file, "r") as psm_decoy_file:
                    logger.info(self.decoy_file)
                    spamreader = csv.DictReader(psm_decoy_file, delimiter="\t")
                    for row in spamreader:
                        row = self.get_alternative_proteins_from_input(row)
                        pdec.append(row)
                all_decoy = pdec

            # Combine the lists
            all_psms = all_target + all_decoy

        elif self.combined_files:
            if isinstance(self.combined_files, (list,)):
                all = []
                for c_files in self.combined_files:
                    c_all = []
                    with open(c_files, "r") as psm_file:
                        logger.info(c_files)
                        spamreader = csv.DictReader(psm_file, delimiter="\t")
                        for row in spamreader:
                            row = self.get_alternative_proteins_from_input(row)
                            c_all.append(row)
                    all = all + c_all
            else:
                c_all = []
                with open(self.combined_files, "r") as psm_file:
                    logger.info(self.combined_files)
                    spamreader = csv.DictReader(psm_file, delimiter="\t")
                    for row in spamreader:
                        row = self.get_alternative_proteins_from_input(row)
                        c_all.append(row)
                all = c_all
            all_psms = all

        elif self.directory:
            all_files = os.listdir(self.directory)
            all = []
            for files in all_files:
                psm_per_file = []
                with open(files, "r") as psm_file:
                    logger.info(files)
                    spamreader = csv.DictReader(psm_file, delimiter="\t")
                    for row in spamreader:
                        row = self.get_alternative_proteins_from_input(row)
                        psm_per_file.append(row)
                all = all + psm_per_file
            all_psms = all

        psms_all_filtered = []
        for psms in all_psms:
            if self.POSTERIOR_ERROR_PROB in psms.keys():
                try:
                    float(psms[self.POSTERIOR_ERROR_PROB])
                    psms_all_filtered.append(psms)
                except ValueError as e:  # noqa F841
                    pass
            else:
                try:
                    float(psms[self.scoring_variable])
                    psms_all_filtered.append(psms)
                except ValueError as e:  # noqa F841
                    pass

        # Filter by pep
        try:
            logger.info("Sorting by {}".format(self.POSTERIOR_ERROR_PROB))
            all_psms = sorted(
                psms_all_filtered,
                key=lambda x: float(x[self.POSTERIOR_ERROR_PROB]),
                reverse=False,
            )
        except KeyError:
            logger.info("Cannot Sort by {} the values do not exist".format(self.POSTERIOR_ERROR_PROB))
            logger.info("Sorting by {}".format(self.scoring_variable))
            if self.parameter_file_object.psm_score_type == Score.ADDITIVE_SCORE_TYPE:
                all_psms = sorted(
                    psms_all_filtered,
                    key=lambda x: float(x[self.scoring_variable]),
                    reverse=True,
                )
            if self.parameter_file_object.psm_score_type == Score.MULTIPLICATIVE_SCORE_TYPE:
                all_psms = sorted(
                    psms_all_filtered,
                    key=lambda x: float(x[self.scoring_variable]),
                    reverse=False,
                )

        list_of_psm_objects = []
        peptide_tracker = set()
        all_sp_proteins = set(self.digest.swiss_prot_protein_set)
        # We only want to get unique peptides... using all messes up scoring...
        # Create Psm objects with the identifier, percscore, qvalue, pepvalue, and possible proteins...

        peptide_to_protein_dictionary = self.digest.peptide_to_protein_dictionary

        initial_poss_prots = []
        logger.info("Number of PSMs in the input data: {}".format(len(all_psms)))
        psms_with_alternative_proteins = self._find_psms_with_alternative_proteins(raw_psms=all_psms)
        logger.info(
            "Number of PSMs that have alternative proteins in the input data {}".format(
                len(psms_with_alternative_proteins)
            )
        )
        if len(psms_with_alternative_proteins) == 0:
            logger.warning(
                "No PSMs in the input have alternative proteins. "
                "Make sure your input is properly formatted. "
                "Alternative Proteins will be retrieved from the fasta database"
            )
        for psm_info in all_psms:
            current_peptide = psm_info[self.PEPTIDE]
            # Define the Psm...
            if current_peptide not in peptide_tracker:
                psm = Psm(identifier=current_peptide)
                # Attempt to add variables from PSM info...
                # If they do not exist in the psm info then we skip...
                try:
                    psm.percscore = float(psm_info[self.SCORE])
                except KeyError:
                    pass
                try:
                    psm.qvalue = float(psm_info[self.Q_VALUE])
                except KeyError:
                    pass
                try:
                    psm.pepvalue = float(psm_info[self.POSTERIOR_ERROR_PROB])
                except KeyError:
                    pass
                # If user has a custom score IE not q-value or pep_value...
                if self.load_custom_score:
                    # Then we look for it...
                    psm.custom_score = float(psm_info[self.scoring_variable])
                psm.possible_proteins = []
                psm.possible_proteins.append(psm_info[self.PROTEIN_IDS])
                psm.possible_proteins = psm.possible_proteins + [x for x in psm_info[self.ALTERNATIVE_PROTEINS] if x]
                # Remove potential Repeats
                if self.parameter_file_object.inference_type != Inference.FIRST_PROTEIN:
                    psm.possible_proteins = sorted(list(set(psm.possible_proteins)))

                input_poss_prots = copy.copy(psm.possible_proteins)

                # Get PSM ID
                psm.psm_id = psm_info[self.PSMID]

                # Split peptide if flanking
                current_peptide = Psm.split_peptide(peptide_string=current_peptide)

                if not current_peptide.isupper() or not current_peptide.isalpha():
                    # If we have mods remove them...
                    peptide_string = current_peptide.upper()
                    stripped_peptide = Psm.remove_peptide_mods(peptide_string)
                    current_peptide = stripped_peptide
                # Add the other possible_proteins from insilicodigest here...
                try:
                    current_alt_proteins = sorted(list(peptide_to_protein_dictionary[current_peptide]))
                except KeyError:
                    current_alt_proteins = []
                    logger.debug(
                        "Peptide {} was not found in the supplied DB for Proteins {}".format(
                            current_peptide, ";".join(psm.possible_proteins)
                        )
                    )
                    for poss_prot in psm.possible_proteins:
                        self.digest.peptide_to_protein_dictionary.setdefault(current_peptide, set()).add(poss_prot)
                        self.digest.protein_to_peptide_dictionary.setdefault(poss_prot, set()).add(current_peptide)
                        logger.debug(
                            "Adding Peptide {} and Protein {} to Digest dictionaries".format(current_peptide, poss_prot)
                        )

                # Sort Alt Proteins by Swissprot then Trembl...
                identifiers_sorted = DataStore.sort_protein_strings(
                    protein_string_list=current_alt_proteins,
                    sp_proteins=all_sp_proteins,
                    decoy_symbol=self.parameter_file_object.decoy_symbol,
                )

                # Restrict to 50 possible proteins
                psm = self._fix_alternative_proteins(
                    append_alt_from_db=self.append_alt_from_db,
                    identifiers_sorted=identifiers_sorted,
                    max_proteins=self.MAX_ALLOWED_ALTERNATIVE_PROTEINS,
                    psm=psm,
                    parameter_file_object=self.parameter_file_object,
                )

                list_of_psm_objects.append(psm)
                peptide_tracker.add(current_peptide)

                initial_poss_prots.append(input_poss_prots)

        self.psms = list_of_psm_objects

        self._check_initial_database_overlap(
            initial_possible_proteins=initial_poss_prots, initial_protein_peptide_map=self.initial_protein_peptide_map
        )

        logger.info("Length of PSM Data: {}".format(len(self.psms)))

        logger.info("Finished GenericReader.read_psms...")

    def _find_psms_with_alternative_proteins(self, raw_psms):

        psms_with_alternative_proteins = [x for x in raw_psms if x["alternative_proteins"]]

        return psms_with_alternative_proteins



import os
import tempfile
from unittest import TestCase

from pkg_resources import resource_filename

import pyproteininference
from pyproteininference import in_silico_digest
from pyproteininference.parameters import ProteinInferenceParameter

TEST_DATABASE = resource_filename("pyproteininference", "../tests/data/test_database.fasta")
TARGET_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_target.txt")
DECOY_FILE = resource_filename("pyproteininference", "../tests/data/test_perc_data_decoy.txt")
PARAMETER_FILE = resource_filename("pyproteininference", "../tests/data/test_params_peptide_centric.yaml")

OUTPUT_DIR = tempfile.gettempdir()


class TestImproperMethodCalls(TestCase):
    def test_incorrect_method_calls(self):

        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        # STEP 1: Load parameter file #
        protein_inference_parameters = ProteinInferenceParameter(yaml_param_filepath=PARAMETER_FILE)

        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        # STEP 2: Start with running an In Silico Digestion #
        digest = in_silico_digest.PyteomicsDigest(
            database_path=TEST_DATABASE,
            digest_type=protein_inference_parameters.digest_type,
            missed_cleavages=protein_inference_parameters.missed_cleavages,
            reviewed_identifier_symbol=protein_inference_parameters.reviewed_identifier_symbol,
            max_peptide_length=protein_inference_parameters.restrict_peptide_length,
            id_splitting=True,
        )
        digest.digest_fasta_database()

        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        # STEP 3: Read PSM Data #
        pep_and_prot_data = pyproteininference.reader.GenericReader(
            target_file=TARGET_FILE,
            decoy_file=DECOY_FILE,
            parameter_file_object=protein_inference_parameters,
            digest=digest,
            append_alt_from_db=False,
        )

        # Validate that we will get a ValueError if we try to load the DataStore object
        # WITHOUT running read_psms from Reader class
        with self.assertRaises(ValueError):
            data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        # Now read the psms
        pep_and_prot_data.read_psms()

        # Now load data object properly...
        data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        # Try to run restrict_psm_data with empty data...
        data.main_data_form = []
        with self.assertRaises(ValueError):
            data.restrict_psm_data()

        # Try to run create_scoring_input with empty data...
        with self.assertRaises(ValueError):
            data.create_scoring_input()

        # Reload data object properly again
        data = pyproteininference.datastore.DataStore(pep_and_prot_data, digest=digest)

        # Run data restrict
        data.restrict_psm_data()

        # Try to run exclude_non_distinguishing_peptides without creating scoring input
        with self.assertRaises(ValueError):
            data.exclude_non_distinguishing_peptides()

        # Try to initialize a Score object without running create_scoring_input
        with self.assertRaises(ValueError):
            score = pyproteininference.scoring.Score(data=data)

        # Finally intiate scoring input properly
        data.create_scoring_input()

        # Try to run picker without scored data...
        with self.assertRaises(ValueError):
            data.protein_picker()

        # Try to run inference without scored data...
        with self.assertRaises(ValueError):
            group = pyproteininference.inference.PeptideCentric(data=data, digest=digest)

        with self.assertRaises(ValueError):
            group = pyproteininference.inference.Parsimony(data=data, digest=digest)

        with self.assertRaises(ValueError):
            group = pyproteininference.inference.Exclusion(data=data, digest=digest)

        with self.assertRaises(ValueError):
            group = pyproteininference.inference.Inclusion(data=data, digest=digest)

        with self.assertRaises(ValueError):
            group = pyproteininference.inference.FirstProtein(data=data, digest=digest)

        # Score psms properly
        score = pyproteininference.scoring.Score(data=data)
        score.score_psms(score_method=protein_inference_parameters.protein_score)

        # Run picker properly
        data.protein_picker()

        # Try to run q value calculation without grouped scored proteins...
        with self.assertRaises(ValueError):
            data.calculate_q_values()

        # Run inference properly
        group = pyproteininference.inference.PeptideCentric(data=data, digest=digest)
        group.infer_proteins()

        # Run q value calculation...
        data.calculate_q_values()

        # Export as normal
        export_type = protein_inference_parameters.export
        export = pyproteininference.export.Export(data=data)
        export.export_to_csv(directory=os.path.join(OUTPUT_DIR, "leads"), export_type=export_type)
